import numpy as np
from projectile import simulator
import mogp_emulator

try:
    import matplotlib.pyplot as plt
    makeplots = True
except ImportError:
    makeplots = False

# define a helper function for making plots

def plot_solution(field, title, filename, simulation_points, validation_points, tri):
    plt.figure(figsize=(4,3))
    plt.tripcolor(validation_points[:,0], validation_points[:,1], tri.triangles,
                  field, vmin=0, vmax=5000.)
    cb = plt.colorbar()
    plt.scatter(simulation_points[:,0], simulation_points[:,1])
    plt.xlabel("log drag coefficient")
    plt.ylabel("Launch velocity (m/s)")
    cb.set_label("Projectile distance (m)")
    plt.title(title)
    plt.tight_layout()
    plt.savefig(filename, dpi=200)
    
# A tutorial illustrating effectiveness of mean functions and priors for GP emulation

# Most often, we are not able to sample very densely from a simulation, so we
# have relatively few samples per input parameter. This can lead to some problems
# when constructing a robust emulator. This tutorial illustrates how we can build
# better emulators using the tools in mogp_emulator.

# We need to draw some samples from the space to run some simulations and build our
# emulators. We use a LHD design with only 6 sample points.

lhd = mogp_emulator.LatinHypercubeDesign([(-4., 0.), (0., 1000.)])

n_simulations = 6
simulation_points = lhd.sample(n_simulations)
simulation_output = np.array([simulator(p) for p in simulation_points])

# Next, fit the surrogate GP model using MLE, zero mean, and no priors.
# Print out hyperparameter values as correlation lengths, sigma, and nugget

gp = mogp_emulator.GaussianProcess(simulation_points, simulation_output)
gp = mogp_emulator.fit_GP_MAP(gp)

print("Zero mean and no priors:")
print("Correlation lengths = {}".format(np.sqrt(np.exp(-gp.theta[:2]))))
print("Sigma = {}".format(np.sqrt(np.exp(gp.theta[2]))))
print("Nugget = {}".format(gp.nugget))
print()

# We can look at how the emulator performs by comparing the emulator output to
# a large number of validation points. Since this simulation is cheap, we can
# actually compute this for a large number of points.

n_valid = 1000
validation_points = lhd.sample(n_valid)
validation_output = np.array([simulator(p) for p in validation_points])

if makeplots:
    import matplotlib.tri
    tri = matplotlib.tri.Triangulation((validation_points[:,0]+4.)/4.,
                                       (validation_points[:,1]/1000.))

    plot_solution(validation_output, "True simulator", "simulator_output.png",
                  simulation_points, validation_points, tri)

# Now predict values with the emulator and plot output and error

predictions = gp.predict(validation_points)

if makeplots:
    plot_solution(predictions.mean, "MLE emulator", "emulator_output_MLE.png",
                  simulation_points, validation_points, tri)

# This is not very good! The simulation points are too sparsely sampled to give the
# emulator any idea what to do about the function shape. We just know the value at a few
# points, and it throws up its hands and predicts zero everywhere else.

# To improve this, we will specify a mean function and some priors to ensure that if we are
# far away from an evaluation point we will still get some information from the emulator.

# We specify the mean function using a string, which follows a similar approach to R-style
# formulas. There is an implicit constant term, and we use x[index] to specify how we
# want the formula to depend on the inputs. We choose a simple linear form here, which has
# three fitting parameters in addition to the correlations lengths, sigma, and nugget
# parameters above.

meanfunc = "x[0]+x[1]"

# We now set priors for all of the hyperparameters to better constrain the estimation procedure.
# We assume normal priors for the mean function parameters with a large variance (to not constrain
# our choice too much). Note that the mean function parameters are on a linear scale, while the
# correlation lengths, sigma, and nugget are on a logarithmic scale. Thus, if we choose normal
# priors on the correlation lengths, these will actually be lognormal distributions.

# Finally, we choose inverse gamma and gamma distributions for the priors on sigma and the nugget
# as those are typical conjugate priors for variances/precisions. We pick them to be where they are as
# we expect sigma to be large (as the function is very sensitive to inputs) while we want the
# nugget to be small.

priors = [mogp_emulator.Priors.NormalPrior(0., 10.),
          mogp_emulator.Priors.NormalPrior(0., 10.),
          mogp_emulator.Priors.NormalPrior(0., 10.),
          mogp_emulator.Priors.NormalPrior(0., 1.),
          mogp_emulator.Priors.NormalPrior(-10., 1.),
          mogp_emulator.Priors.InvGammaPrior(1., 1.),
          mogp_emulator.Priors.GammaPrior(1., 1.)]

# Now, construct another GP using the mean function and priors. note that we also specify that we
# want to estimate the nugget based on our prior, rather than adaptively fitting it as we did in
# the first go.

gp_map = mogp_emulator.GaussianProcess(simulation_points, simulation_output,
                                       mean=meanfunc, priors=priors, nugget="fit")
gp_map = mogp_emulator.fit_GP_MAP(gp_map)

print("With mean and priors:")
print("Mean function parameters = {}".format(gp_map.theta[:3]))
print("Correlation lengths = {}".format(np.sqrt(np.exp(-gp_map.theta[3:5]))))
print("Sigma = {}".format(np.sqrt(np.exp(gp_map.theta[-2]))))
print("Nugget = {}".format(gp_map.nugget))

# Use the new fit GP to predict the validation points and plot to see if this improved
# the fit to the true data:

predictions_map = gp_map.predict(validation_points)

if makeplots:
    plot_solution(predictions_map.mean, "Mean/Prior emulator", "emulator_output_MAP.png",
                  simulation_points, validation_points, tri)



'''
This benchmark performs convergence tests using the MICE experimental design applied to the
2D Branin function. Details of the 2D Branin function can be found at
https://www.sfu.ca/~ssurjano/branin.html. The code samples the Branin function using the
MICE experimental design algorithm with a varying number of target points, and then tests
the convergence of the resulting emulators. As the number of targe points increases, the
prediction error and prediction variance should decrease.

(Note however that eventually, the predictions worsen once the number of target points becomes
large enough that the points become too densely sampled. In this case, the points become
co-linear and the resulting covariance matrix is singular and cannot be inverted. To avoid
this problem, the code iteratively adds additional noise to the covariance function to
stabilize the inversion. However, this noise reduces the accuracy of the predictions. The
values chosen for this benchmark attempt to avoid this, but in some cases this still becomes
a problem due to the inherent smoothness of the squared exponential covariance function.)
'''

import numpy as np
from mogp_emulator import GaussianProcess, fit_GP_MAP
from mogp_emulator import MICEDesign, MonteCarloDesign, LatinHypercubeDesign
from scipy.stats import uniform
try:
    import matplotlib.pyplot as plt
    makeplots = True
except ImportError:
    makeplots = False

problem = 'branin'

def branin_2d(x):
    "2D Branin function, see https://www.sfu.ca/~ssurjano/branin.html for more information"
    if np.array(x).shape == (2,):
        x1, x2 = x
    else:
        assert len(np.array(x).shape) == 2
        assert np.array(x).shape[1] == 2
        x1 = x[:,0]
        x2 = x[:,1]
    a, b, c, r, s, t = 1., 5.1/4./np.pi**2, 5./np.pi, 6., 10., 1./8./np.pi
    return a*(x2 - b*x1**2 + c*x1 - r)**2 + s*(1. - t)*np.cos(x1) + s

def oscillatory_4d(x):
    "4d oscillatory function for testing"
    if np.array(x).shape == (4,):
        x1, x2, x3, x4 = x
    else:
        assert len(np.array(x).shape) == 2
        assert np.array(x).shape[1] == 4
        x1 = x[:,0]
        x2 = x[:,1]
        x3 = x[:,2]
        x4 = x[:,3]
    a, b, c, d, w = 1.85, 2.51, 1.94, 2.7, 0.43
    return np.cos(a*x1 + b*x2 + c*x3 + d*x4 + 2.*np.pi*w)

if problem == 'branin':
    f = branin_2d
    n_dim = 2
    design_space = [uniform(loc = -5., scale = 15.).ppf, uniform(loc = 0., scale = 15.).ppf]
    simulations = [5, 10, 15, 20, 25, 30]
else:
    f = oscillatory_4d
    n_dim = 4
    design_space = 4
    simulations = [5, 10, 15, 20, 25, 30, 35]

def generate_input_data(n_simulations, method = "random"):
    "Generate random points x1 and x2 for evaluating the multivalued 2D Branin function"

    n_simulations = int(n_simulations)
    assert(n_simulations > 0)
    assert method == "random" or method == "lhd"

    if method == "random":
        ed = MonteCarloDesign(design_space)
    elif method == "lhd":
        ed = LatinHypercubeDesign(design_space)
    inputs = ed.sample(n_simulations)
    return inputs

def generate_target_data(inputs):
    "Generate target data for multivalued emulator benchmark"

    inputs = np.array(inputs)
    assert len(inputs.shape) == 2
    assert inputs.shape[1] == n_dim
    n_simulations = inputs.shape[0]

    targets = f(inputs)

    return targets

def generate_training_data(n_simulations):
    "Generate n_simulations input data and evaluate using n_emulators different parameter values"

    inputs = generate_input_data(n_simulations, method = "lhd")
    targets = generate_target_data(inputs)

    return inputs, targets

def generate_test_data(n_testing):
    "Generate n_testing points for testing the accuracy of an emulator"

    testing = generate_input_data(n_testing, method = "random")
    test_targets = generate_target_data(testing)

    return testing, test_targets

def run_model(n_simulations, n_testing):
    "Generate training data, fit emulator, and test model accuracy on random points, returning RMSE"

    # run MICE Model
    print('running MICE')
    ed = LatinHypercubeDesign(design_space)

    n_init = 5

    md = MICEDesign(ed, f, n_samples = n_simulations - n_init, n_init = n_init, n_cand = 100)

    md.run_sequential_design()

    inputs_mice = md.get_inputs()
    targets_mice = md.get_targets()

    print('fitting GPs')

    gp_mice = GaussianProcess(inputs_mice, np.squeeze(targets_mice))
    gp_mice = fit_GP_MAP(gp_mice)

    # run LHD model
    inputs, targets = generate_training_data(n_simulations)

    gp = GaussianProcess(inputs, targets)
    gp = fit_GP_MAP(gp)

    print("making predictions")

    testing, test_targets = generate_test_data(n_testing)

    norm_const = np.max(test_targets)-np.min(test_targets)

    test_vals_mice, unc_mice, deriv = gp_mice.predict(testing, deriv = False, unc = True)
    test_vals, unc, deriv = gp.predict(testing, deriv = False, unc = True)

    return (np.sqrt(np.sum((test_vals - test_targets)**2)/float(n_testing))/norm_const,
            np.sqrt(np.sum(unc**2)/float(n_testing))/norm_const**2,
            np.sqrt(np.sum((test_vals_mice - test_targets)**2)/float(n_testing))/norm_const,
            np.sqrt(np.sum(unc_mice**2)/float(n_testing))/norm_const**2)

def plot_model_errors(simulation_list, error, unc, error_mice, unc_mice, n_testing):
    "Makes plot showing accuracy of emulator as a function of n_simulations"

    plt.figure(figsize=(4,3))
    plt.semilogy(simulation_list, error_mice,'-o', label = 'MICE')
    plt.semilogy(simulation_list, error,'-x', label = 'LHD')
    plt.xlabel('Number of design points')
    plt.ylabel('Average prediction RMSE')
    plt.legend()
    plt.title('Error for '+str(n_testing)+' predictions')
    plt.savefig('mice_'+problem+'_error.png',bbox_inches='tight')

    plt.figure(figsize=(4,3))
    plt.semilogy(simulation_list, unc_mice,'-o', label = "MICE")
    plt.semilogy(simulation_list, unc,'-x', label = 'LHD')
    plt.xlabel('Number of design points')
    plt.ylabel('Average prediction variance')
    plt.legend()
    plt.title('Uncertainty for '+str(n_testing)+' predictions')
    plt.savefig('mice_'+problem+'_unc.png',bbox_inches='tight')

def run_all_models(n_testing, simulation_list, n_iter = 10):
    "Runs all models, printing out results and optionally making plots"

    n_simtrials = len(simulation_list)

    errors = np.zeros((n_simtrials, n_iter))
    uncs = np.zeros((n_simtrials, n_iter))
    errors_mice = np.zeros((n_simtrials, n_iter))
    uncs_mice = np.zeros((n_simtrials, n_iter))

    for iteration in range(n_iter):
        for sim_index in range(n_simtrials):
            print(sim_index, iteration)
            (errors[sim_index, iteration],
             uncs[sim_index, iteration],
             errors_mice[sim_index, iteration],
             uncs_mice[sim_index, iteration]) = run_model(simulation_list[sim_index], n_testing)

    error = np.mean(errors, axis = -1)
    unc = np.mean(uncs, axis = -1)
    error_mice = np.mean(errors_mice, axis = -1)
    unc_mice = np.mean(uncs_mice, axis = -1)

    print("\n")
    print("Convergence test results:")
    print("Num. design points  RMSE LHD            RMSE MICE")
    for sim, err, mice_err in zip(simulation_list, error, error_mice):
        print('{:19} {:19} {:19}'.format(str(sim), str(err), str(mice_err)))

    print("\n")
    print("Num. design points  Variance LHD        Variance MICE")
    for sim, un, mice_un in zip(simulation_list, unc, unc_mice):
        print('{:19} {:19} {:19}'.format(str(sim), str(un), str(mice_un)))

    if makeplots:
        plot_model_errors(simulation_list, error, unc, error_mice, unc_mice, n_testing)

if __name__ == '__main__':
    run_all_models(100, [int(x) for x in simulations], n_iter = 10)


import numpy as np
from numpy.testing import assert_allclose
import scipy.stats
import pytest
from inspect import signature
import types
from ..ExperimentalDesign import ExperimentalDesign, MonteCarloDesign, LatinHypercubeDesign

def test_ExperimentalDesign_init():
    "test the init method of ExperimentalDesign"
    
    ed = ExperimentalDesign(3)
    assert ed.n_parameters == 3
    assert len(ed.distributions) == 3
    for dist in ed.distributions:
        assert callable(dist)
        assert len(signature(dist).parameters) == 1
    
    ed = ExperimentalDesign(3.)
    assert ed.n_parameters == 3
    assert len(ed.distributions) == 3
    for dist in ed.distributions:
        assert callable(dist)
        assert len(signature(dist).parameters) == 1
    
    ed = ExperimentalDesign(3, (0., 1.))
    assert ed.n_parameters == 3
    assert len(ed.distributions) == 3
    for dist in ed.distributions:
        assert callable(dist)
        assert len(signature(dist).parameters) == 1
    
    ed = ExperimentalDesign(3, scipy.stats.uniform(loc = 0., scale = 1.).ppf)
    assert ed.n_parameters == 3
    assert len(ed.distributions) == 3
    for dist in ed.distributions:
        assert callable(dist)
        assert len(signature(dist).parameters) == 1

    ed = ExperimentalDesign(2, [(0., 5.), (0., 10.)])
    assert ed.n_parameters == 2
    assert len(ed.distributions) == 2
    for dist in ed.distributions:
        assert callable(dist)
        assert len(signature(dist).parameters) == 1
    
    ed = ExperimentalDesign([(0., 1.), (0., 5.), (0., 10.)])
    assert ed.n_parameters == 3
    assert len(ed.distributions) == 3
    for dist in ed.distributions:
        assert callable(dist)
        assert len(signature(dist).parameters) == 1
    
    ed = ExperimentalDesign([scipy.stats.uniform(loc = 0., scale = 1.).ppf,
                             scipy.stats.uniform(loc = 0., scale = 5.).ppf,
                             scipy.stats.uniform(loc = 0., scale = 10.).ppf])
    assert ed.n_parameters == 3
    assert len(ed.distributions) == 3
    for dist in ed.distributions:
        assert callable(dist)
        assert len(signature(dist).parameters) == 1
    
    ed = ExperimentalDesign([(0., 1.), scipy.stats.uniform(loc = 0., scale = 5.).ppf, (0., 10.)])
    assert ed.n_parameters == 3
    assert len(ed.distributions) == 3
    for dist in ed.distributions:
        assert callable(dist)
        assert len(signature(dist).parameters) == 1
    
def test_ExperimentalDesign_init_failures():
    "test situations where init for ExperimentalDesign should fail"
    
    with pytest.raises(ValueError):
        ed = ExperimentalDesign(-1)

    with pytest.raises(ValueError):
        ed = ExperimentalDesign(4, (0., 2., 2.))

    with pytest.raises(TypeError):
        ed = ExperimentalDesign(3, (0., 2., 2.))
    
    with pytest.raises(ValueError):
        ed = ExperimentalDesign(3, scipy.stats.uniform.ppf)
        
    with pytest.raises(ValueError):
        ed = ExperimentalDesign(2, [(0., 1.), (0., 5.), (0., 10.)])
        
    with pytest.raises(ValueError):
        ed = ExperimentalDesign([(1., 0.), (0., 5.), (0., 10.)])
        
    with pytest.raises(ValueError):
        ed = ExperimentalDesign(2, [(0., 1., 2), (0., 5.)])
        
    with pytest.raises(ValueError):
        ed = ExperimentalDesign([scipy.stats.uniform.ppf, (0., 5.)])
        
    
def test_ExperimentalDesign_get_n_parameters():
    "test get_n_samples method of ExperimentalDesign"
    
    ed = ExperimentalDesign(3)
    assert ed.get_n_parameters() == 3
    
def test_ExperimentalDesign_get_method():
    "test get_method method of ExperimentalDesign"
    
    ed = ExperimentalDesign(3)
    with pytest.raises(NotImplementedError):
        ed.get_method()
        
def test_ExperimentalDesign_draw_samples():
    "test _draw_samples method of ExperimentalDesign"
    
    ed = ExperimentalDesign(3)
    with pytest.raises(NotImplementedError):
        ed.sample(3)
        
def test_ExperimentalDesign_sample():
    "test sample method of ExperimentalDesign"

    ed = ExperimentalDesign(3)
    
    def tmp_draw_samples(self, n_samples):
        return np.ones((n_samples, self.get_n_parameters()))*0.5

    ed._draw_samples = types.MethodType(tmp_draw_samples, ed)
    
    sample = ed.sample(2)
    sample_expected = np.ones((2, 3))*0.5
    assert_allclose(sample, sample_expected)
    
    def tmp_draw_samples(self, n_samples):
        return np.ones((n_samples, self.get_n_parameters()))*0.25

    ed._draw_samples = types.MethodType(tmp_draw_samples, ed)
    
    sample = ed.sample(2)
    sample_expected = np.ones((2, 3))*0.25
    assert_allclose(sample, sample_expected)

def test_ExperimentalDesign_sample_failures():
    "test situation where ExperimentalDesign should raise an exception"
    
    with pytest.raises(AssertionError):
        ed = ExperimentalDesign(3)
        sample = ed.sample(-2)
    
    with pytest.raises(AssertionError):
        f = lambda x: np.nan
        ed = ExperimentalDesign(2, f)
        
        def tmp_draw_samples(self, n_samples):
            return np.ones((n_samples, self.get_n_parameters()))*0.5

        ed._draw_samples = types.MethodType(tmp_draw_samples, ed)
        sample = ed.sample(2)

def test_ExperimentalDesign_str():
    "test string method of ExperimentalDesign"
    
    ed = ExperimentalDesign(3)
    assert str(ed) == "Experimental Design with 3 parameters"

def test_MonteCarloDesign_init():
    "test that init method correctly sets method for Monte Carlo"
    
    ed = MonteCarloDesign(3)
    assert ed.get_method() == "Monte Carlo"
    assert str(ed) == "Monte Carlo Experimental Design with 3 parameters"
    
def test_MonteCarloDesign_draw_samples():
    "test low level method for drawing samples from a monte carlo design"
    
    ed = MonteCarloDesign(3)
    np.random.seed(73859)
    sample = ed._draw_samples(4)
    sample_expected = np.array([[0.4587231420553423, 0.8712779046673758, 0.9592574641762792],
                                [0.6863958521075596, 0.1642774978932529, 0.5102796498909351],
                                [0.0945231904820333, 0.7578914914795801, 0.4338919363034113],
                                [0.8612216997446789, 0.0194411385272029, 0.0765283763808666]])
    assert_allclose(sample, sample_expected)
    
def test_MonteCarloDesign_sample():
    "test method to samples from a Monte Carlo design"
    
    ed = MonteCarloDesign(3)
    np.random.seed(73859)
    sample = ed.sample(4)
    sample_expected = np.array([[0.4587231420553423, 0.8712779046673758, 0.9592574641762792],
                                [0.6863958521075596, 0.1642774978932529, 0.5102796498909351],
                                [0.0945231904820333, 0.7578914914795801, 0.4338919363034113],
                                [0.8612216997446789, 0.0194411385272029, 0.0765283763808666]])
    assert_allclose(sample, sample_expected)
    
    ed = MonteCarloDesign([scipy.stats.uniform(loc = 0., scale = 5.).ppf, scipy.stats.uniform(loc = 0., scale = 2.).ppf])
    sample = ed.sample(3)
    sample_expected = np.array([[1.4976951596996646, 1.9820310379261286],
                                [4.841151376332333 , 0.9447917510829495],
                                [3.8191224939802453, 0.6532873534667354]])
    assert_allclose(sample, sample_expected)
        
def test_LatinHypercubeDesign_init():
    "test that init method correctly sets method for Latin Hypercube"
    
    ed = LatinHypercubeDesign(3)
    assert ed.get_method() == "Latin Hypercube"
    assert str(ed) == "Latin Hypercube Experimental Design with 3 parameters"
    
def test_LatinHypercubeDesign_draw_samples():
    "test low level method for drawing samples from a latin hypercube design"
    
    ed = LatinHypercubeDesign(2)
    np.random.seed(73859)
    sample = ed._draw_samples(5)
    sample_expected = np.array([[0.3057514377091521, 0.2505993516661572],
                                [0.5739619489876799, 0.0113841212300836],
                                [0.7776329187470997, 0.5688057959369438],
                                [0.8885708869643932, 0.7610737581544206],
                                [0.0628594487771623, 0.8735926821354483]])
    assert_allclose(sample, sample_expected)
    
def test_LatinHypercubeDesign_sample():
    "test method to samples from a latin hypercube design"
    
    ed = LatinHypercubeDesign(2)
    np.random.seed(73859)
    sample = ed.sample(5)
    sample_expected = np.array([[0.3057514377091521, 0.2505993516661572],
                                [0.5739619489876799, 0.0113841212300836],
                                [0.7776329187470997, 0.5688057959369438],
                                [0.8885708869643932, 0.7610737581544206],
                                [0.0628594487771623, 0.8735926821354483]])
    assert_allclose(sample, sample_expected)
    
    ed = LatinHypercubeDesign([scipy.stats.uniform(loc = 0., scale = 5.).ppf, scipy.stats.uniform(loc = 0., scale = 2.).ppf])
    sample = ed.sample(3)
    sample_expected = np.array([[3.0319074209362724, 1.4378127211566698],
                                [3.8539110146362914, 0.186886121478102 ],
                                [0.3690344832370781, 1.1036076298948434]])
    assert_allclose(sample, sample_expected)


"""Dimension reduction benchmark (gKDR)

Plot L_1 model error against reduced dimension, for a one-hundred
dimensional problem.  The model is a Gaussian process (with
maximum-likelihood hyperparameters), with inputs determined by gKDR,
varying structural dimension and with a selection of scale parameters.
The inputs to the model are 100 points chosen uniformly at random in
100 dimensions, and the observations are a linear mapping of these
into one dimension.
"""

import os
from contextlib import redirect_stdout
import numpy as np
from mogp_emulator import fit_GP_MAP
from mogp_emulator import DimensionReduction
from mogp_emulator import gKDR
try:
    import matplotlib.pyplot as plt
    have_plt = True
except ImportError:
    print("matplotlib.pyplot import error: skipping plots")
    have_plt = False

dev_null = open(os.devnull, "w")

def fn(x):
    return x[0]


def make_plot(loss):
    """Produce a plot of loss as a function of structural dimension."""
    plt.xlabel("Structural dimension")
    plt.ylabel("Loss ($L_1$)")
    for ic in range(loss.shape[0]):
        assert(np.all(loss[ic,:,2:] == loss[ic,0,2:]))
        cX, cY = loss[ic,0,2:]
        plt.plot(loss[ic,:,0], loss[ic,:,1], 'x-', label = "$c_X = {}, c_Y = {}$".format(cX, cY))

    plt.ylim(bottom=0.0)
    plt.legend()
    # plt.show()
    plt.savefig("benchmark_kdr_GP_loss.pdf")


def run():
    """Run the benchmark"""
    N = 100
    D = 100
    np.random.seed(3)
    X = np.random.random((N, D))
    Y = np.apply_along_axis(fn, 1, X)

    def compute_loss(k):
        ## ignore occasional warning messages from the GP fitting
        with redirect_stdout(dev_null):
            loss = gKDR._compute_loss(X, Y, fit_GP_MAP,
                                      cross_validation_folds = 5,
                                      K = k, X_scale = cX, Y_scale = cY)

        print("loss(K={}, cX={}, cY={}) = {}".format(k, cX, cY, loss))
        return loss

    K = np.arange(1,101)
    scale_params = [(1.0, 1.0), (2.0, 2.0), (5.0, 5.0)]

    loss = np.zeros((len(scale_params), len(K), 4))

    for ic in range(len(scale_params)):
        cX, cY = scale_params[ic]
        loss[ic,:,:] = np.array([[k, compute_loss(k), cX, cY] for k in K])

    np.save("benchmark_kdr_GP_loss", loss)

    if have_plt:
        make_plot(loss)


if __name__ == '__main__':
    run()



from multiprocessing import Pool
import platform
import numpy as np
from mogp_emulator.GaussianProcess import GaussianProcess, PredictResult
from mogp_emulator.MeanFunction import MeanBase
from mogp_emulator.Kernel import Kernel, SquaredExponential, Matern52
from mogp_emulator.Priors import Prior

class MultiOutputGP(object):
    """
    Implementation of a multiple-output Gaussian Process Emulator.

    Essentially a parallelized wrapper for the predict method. To fit in parallel, use the fit_GP_MAP
    routine

    Required arguments are ``inputs`` and ``targets``, both of which must be numpy arrays. ``inputs``
    can be 1D or 2D (if 1D, assumes second axis has length 1). ``targets`` can be 1D or 2D (if 2D,
    assumes a single emulator and the first axis has length 1).

    Optional arguments specify how each individual emulator is constructed, including the mean
    function, kernel, priors, and how to handle the nugget. Each argument can take values allowed
    by the base ``GaussianProcess`` class, in which case all emulators are assumed to use the
    same value. Any of these arguments can alternatively be a list of values with length matching
    the number of emulators to set those values individually.

    Additional keyword arguments include ``inputdict``, and ``use_patsy``, which control how strings
    are parsed to mean functions, if using.
    """

    def __init__(self, inputs, targets, mean=None, kernel=SquaredExponential(), priors=None,
                 nugget="adaptive", inputdict={}, use_patsy=True):
        """
        Create a new multi-output GP Emulator
        """

        # check input types and shapes, reshape as appropriate for the case of a single emulator
        inputs = np.array(inputs)
        targets = np.array(targets)
        if len(inputs.shape) == 1:
            inputs = np.reshape(inputs, (-1, 1))
        if len(targets.shape) == 1:
            targets = np.reshape(targets, (1, -1))
        elif not (len(targets.shape) == 2):
            raise ValueError("targets must be either a 1D or 2D array")
        if not (len(inputs.shape) == 2):
            raise ValueError("inputs must be either a 1D or 2D array")
        if not (inputs.shape[0] == targets.shape[1]):
            raise ValueError("the first dimension of inputs must be the same length as the second dimension of targets (or first if targets is 1D))")

        self.n_emulators = targets.shape[0]
        self.n = inputs.shape[0]
        self.D = inputs.shape[1]

        if mean is None or isinstance(mean, str) or issubclass(type(mean), MeanBase):
            mean = self.n_emulators*[mean]

        assert isinstance(mean, list), "mean must be None, a string, a mean function, or a list of None/string/mean functions"
        assert len(mean) == self.n_emulators

        if isinstance(kernel, str):
            if kernel == "SquaredExponential":
                kernel = SquaredExponential()
            elif kernel == "Matern52":
                kernel = Matern52()
            else:
                raise ValueError("provided kernel '{}' not a supported kernel type".format(kernel))
        if issubclass(type(kernel), Kernel):
            kernel = self.n_emulators*[kernel]

        assert isinstance(kernel, list), "kernel must be a Kernal subclass or a list of Kernel subclasses"
        assert len(kernel) == self.n_emulators

        if priors is None:
            priors = []
        assert isinstance(priors, list), "priors must be a list of lists of Priors/None"

        if len(priors) == 0:
            priors = self.n_emulators*[[]]

        if not isinstance(priors[0], list):
            priors = self.n_emulators*[priors]

        assert len(priors) == self.n_emulators

        if isinstance(nugget, (str, float)):
            nugget = self.n_emulators*[nugget]

        assert isinstance(nugget, list), "nugget must be a string, float, or a list of strings and floats"
        assert len(nugget) == self.n_emulators

        self.emulators = [ GaussianProcess(inputs, single_target, m, k, p, n, inputdict, use_patsy)
                           for (single_target, m, k, p, n) in zip(targets, mean, kernel, priors, nugget)]


    def predict(self, testing, unc=True, deriv=True, include_nugget=True, processes=None):
        """
        Make a prediction for a set of input vectors

        Makes predictions for each of the emulators on a given set of input vectors. The
        input vectors must be passed as a ``(n_predict, D)`` or ``(D,)`` shaped array-like
        object, where ``n_predict`` is the number of different prediction points under
        consideration and ``D`` is the number of inputs to the emulator. If the prediction
        inputs array has shape ``(D,)``, then the method assumes ``n_predict == 1``.
        The prediction points are passed to each emulator and the predictions are collected
        into an ``(n_emulators, n_predict)`` shaped numpy array as the first return value
        from the method.

        Optionally, the emulator can also calculate the uncertainties in the predictions
        (as a variance) and the derivatives with respect to each input parameter. If the
        uncertainties are computed, they are returned as the second output from the method
        as an ``(n_emulators, n_predict)`` shaped numpy array. If the derivatives are
        computed, they are returned as the third output from the method as an
        ``(n_emulators, n_predict, D)`` shaped numpy array. Finally, if uncertainties
        are computed, the ``include_nugget`` flag determines if the uncertainties should
        include the nugget. By default, this is set to ``True``.

        As with the fitting, this computation can be done independently for each emulator
        and thus can be done in parallel.

        :param testing: Array-like object holding the points where predictions will be made.
                        Must have shape ``(n_predict, D)`` or ``(D,)`` (for a single prediction)
        :type testing: ndarray
        :param unc: (optional) Flag indicating if the uncertainties are to be computed.
                    If ``False`` the method returns ``None`` in place of the uncertainty
                    array. Default value is ``True``.
        :type unc: bool
        :param deriv: (optional) Flag indicating if the derivatives are to be computed.
                      If ``False`` the method returns ``None`` in place of the derivative
                      array. Default value is ``True``.
        :type deriv: bool
        :param include_nugget: (optional) Flag indicating if the nugget should be included
                               in the predictive variance. Only relevant if ``unc = True``.
                               Default is ``True``.
        :type include_nugget: bool
        :param processes: (optional) Number of processes to use when making the predictions.
                          Must be a positive integer or ``None`` to use the number of
                          processors on the computer (default is ``None``)
        :type processes: int or None
        :returns: Tuple of numpy arrays holding the predictions, uncertainties, and derivatives,
                  respectively. Predictions and uncertainties have shape ``(n_emulators, n_predict)``
                  while the derivatives have shape ``(n_emulators, n_predict, D)``. If
                  the ``do_unc`` or ``do_deriv`` flags are set to ``False``, then those arrays
                  are replaced by ``None``.
        :rtype: tuple
        """

        testing = np.array(testing)
        if testing.shape == (self.D,):
            testing = np.reshape(testing, (1, self.D))
        assert len(testing.shape) == 2, "testing must be a 2D array"
        assert testing.shape[1] == self.D, "second dimension of testing must be the same as the number of input parameters"
        if not processes is None:
            processes = int(processes)
            assert processes > 0, "number of processes must be a positive integer"

        if platform.system() == "Windows":
            predict_vals = [GaussianProcess.predict(gp, testing, unc, deriv, include_nugget) for gp in self.emulators]
        else:
            with Pool(processes) as p:
                predict_vals = p.starmap(GaussianProcess.predict, [(gp, testing, unc, deriv, include_nugget) for gp in self.emulators])

        # repackage predictions into numpy arrays

        predict_unpacked, unc_unpacked, deriv_unpacked = [np.array(t) for t in zip(*predict_vals)]

        if not unc:
            unc_unpacked = None
        if not deriv:
            deriv_unpacked = None

        return PredictResult(mean=predict_unpacked, unc=unc_unpacked, deriv=deriv_unpacked)

    def __call__(self, testing):
        """
        Interface to predict means by calling the object

        A MultiOutputGP object is callable, which makes predictions of the mean only
        for a given set of inputs. Works similarly to the same method of the base GP
        class. Predictions are made in parallel using the number of available processors.
        """

        return self.predict(testing, unc=False, deriv=False, processes=None)[0]

    def __str__(self):
        """
        Returns a string representation of the model

        :returns: A string representation of the model (indicates number of sub-components
                  and array shapes)
        :rtype: str
        """

        return ("Multi-Output Gaussian Process with:\n"+
                 str(self.n_emulators)+" emulators\n"+
                 str(self.n)+" training examples\n"+
                 str(self.D)+" input variables")




r"""
Kernel module, implements a few standard stationary kernels for use with the
``GaussianProcess`` class. At present, kernels can only be selected manually by setting
the ``kernel`` attribute of the GP. The default is to use the ``SquaredExponential``
kernel, but this can be changed once the ``GaussianProcess`` instance is created.
"""

import numpy as np
from scipy.spatial.distance import cdist

class Kernel(object):
    r"""
    Generic class representing a stationary kernel

    This base class implements the necessary scaffolding for defining a stationary kernel.
    Stationary kernels are only dependent on a distance measure between any two points, so
    the base class holds all the necessary information for doing the distance computation.
    Individual subclasses will implement the functional dependence of the kernel on the
    distance, plus first and second derivatives (if desired) to compute the gradient or
    Hessian of the kernel with respect to the hyperparameters.

    This implementation uses a scaled euclidean distance metric. Each individual parameter
    has a hyperparameter scale associated with it that is used in the distance computation.
    If a different metric is to be defined, a new base class needs to be defined that
    implements the ``calc_r``, and optionally ``calc_drdtheta`` and ``calc_d2rdtheta2``
    methods if gradient or Hessian computation is desired. The methods ``kernel_f``,
    ``kernel_gradient``, and ``kernel_hessian`` can then be used to compute the appropriate
    quantities with no further modification.

    Note that the Kernel object just collates all of the methods together; the class itself
    does not hold any information on the data point or hyperparamters, which are passed
    directly to the appropriate methods. Thus, no information needs to be provided when
    creating a new ``Kernal`` instance.
    """
    def __str__(self):
        r"""
        Defines a string representation of the kernel

        Returns a string representation of the kernel. Note that since the kernel just
        collects methods for kernel evaluations together with no data, this is just a basic
        string that will not change for different instances of the class.

        :returns: String representation of the kernel
        :rtype: str
        """
        return "Stationary Kernel"

    def _check_inputs(self, x1, x2, params):
        r"""
        Common function for checking dimensions of inputs

        This function checks the inputs to any kernel evaluation for consistency and ensures
        that all input arrays have the correct dimensionality. It returns the reformatted
        arrays, the number of inputs, and the number of hyperparameters. If the method
        determines that the array dimensions are not all consistent with one another,
        it will raise an ``AssertionError``. This method is called internally whenever
        the kernel is evaluated.

        :param x1: First parameter array. Should be a 1-D or 2-D array (1-D is acceptable
                   if either there is only a single point, or each point has only a single
                   parameter). If there is more than one parameter, the last dimension
                   must match the last dimension of ``x2`` and be one less than the length
                   of ``params``.
        :type x1: array-like
        :param x2: Second parameter array. The same restrictions apply that hold for ``x1``
                   described above.
        :type x2: array-like
        :param params: Hyperparameter array. Must have a length that is one more than the
                       last dimension of ``x1`` and ``x2``, hence minimum length is 2.
        :type params: array-like
        :returns: A tuple containing the following: reformatted ``x1``, ``n1``, reformatted
                  ``x2``, ``n2``, ``params``, and ``D``. ``x1`` will be an array with
                  dimensions ``(n1, D - 1)``, ``x2`` will be an array with dimensions
                  ``(n2, D - 1)``, and ``params`` will be an array with dimensions ``(D,)``.
                  ``n1``, ``n2``, and ``D`` will be integers.
        """

        params = np.array(params)
        assert params.ndim == 1, "parameters must be a vector"
        D = len(params)
        assert D >= 2, "minimum number of parameters in a covariance kernel is 2"

        x1 = np.array(x1)

        assert x1.ndim == 1 or x1.ndim == 2, "bad number of dimensions in input x1"

        if x1.ndim == 2:
            assert x1.shape[1] == D - 1, "bad shape for x1"
        else:
            if D == 2:
                x1 = np.reshape(x1, (len(x1), 1))
            else:
                x1 = np.reshape(x1, (1, D - 1))

        n1 = x1.shape[0]

        x2 = np.array(x2)

        assert x2.ndim == 1 or x2.ndim == 2, "bad number of dimensions in input x2"

        if x2.ndim == 2:
            assert x2.shape[1] == D - 1, "bad shape for x2"
        else:
            if D == 2:
                x2 = np.reshape(x2, (len(x2), 1))
            else:
                x2 = np.reshape(x2, (1, D - 1))

        n2 = x2.shape[0]

        return x1, n1, x2, n2, params, D


    def calc_r(self, x1, x2, params):
        r"""
        Calculate distance between all pairs of points

        This method computes the scaled Euclidean distance between all pairs of points
        in ``x1`` and ``x2``. Each component distance is multiplied by the corresponding
        hyperparameter prior to summing and taking the square root. For example, if
        ``x1 = [1.]``, ``x2`` = [2.], and ``params = [2., 2.]`` then ``calc_r`` would
        return :math:`{\sqrt{2(1 - 2)^2}=\sqrt{2}}` as an array with shape ``(1,1)``.

        :param x1: First input array. Must be a 1-D or 2-D array, with the length of
                   the last dimension matching the last dimension of ``x2`` and
                   one less than the length of ``params``. ``x1`` may be 1-D if either
                   each point consists of a single parameter (and ``params`` has length
                   2) or the array only contains a single point (in which case, the array
                   will be reshaped to ``(1, D - 1)``).
        :type x1: array-like
        :param x2: Second input array. The same restrictions that apply to ``x1`` also
                   apply here.
        :type x2: array-like
        :param params: Hyperparameter array. Must be 1-D with length one greater than
                       the last dimension of ``x1`` and ``x2``.
        :type params: array-like
        :returns: Array holding all pair-wise distances between points in arrays ``x1``
                  and ``x2``. Will be an array with shape ``(n1, n2)``, where ``n1``
                  is the length of the first axis of ``x1`` and ``n2`` is the length
                  of the first axis of ``x2``.
        :rtype: ndarray
        """

        x1, n1, x2, n2, params, D = self._check_inputs(x1, x2, params)

        exp_theta = np.exp(-params[:(D - 1)])

        r_matrix = cdist(x1, x2, "seuclidean", V = exp_theta)

        if np.any(np.isnan(r_matrix)):
            raise FloatingPointError("NaN enountered in kernel distance computation")

        return r_matrix

    def calc_drdtheta(self, x1, x2, params):
        r"""
        Calculate the first derivative of the distance between all pairs of points with
        respect to the hyperparameters

        This method computes the derivative of the scaled Euclidean distance between
        all pairs of points in ``x1`` and ``x2`` with respect to the hyperparameters.
        The gradient is held in an array with shape ``(D, n1, n2)``, where ``D`` is
        the length of ``params``, ``n1`` is the length of the first axis of ``x1``,
        and ``n2`` is the length of the first axis of ``x2``. This is used in the
        computation of the gradient and Hessian of the kernel. The first index
        represents the different derivatives with respect to each hyperparameter.

        :param x1: First input array. Must be a 1-D or 2-D array, with the length of
                   the last dimension matching the last dimension of ``x2`` and
                   one less than the length of ``params``. ``x1`` may be 1-D if either
                   each point consists of a single parameter (and ``params`` has length
                   2) or the array only contains a single point (in which case, the array
                   will be reshaped to ``(1, D - 1)``).
        :type x1: array-like
        :param x2: Second input array. The same restrictions that apply to ``x1`` also
                   apply here.
        :type x2: array-like
        :param params: Hyperparameter array. Must be 1-D with length one greater than
                       the last dimension of ``x1`` and ``x2``.
        :type params: array-like
        :returns: Array holding the derivative of the pair-wise distances between
                  points in arrays ``x1`` and ``x2`` with respect to the hyperparameters.
                  Will be an array with shape ``(D, n1, n2)``, where ``D`` is the length
                  of ``params``, ``n1`` is the length of the first axis of ``x1`` and
                  ``n2`` is the length of the first axis of ``x2``. The first axis
                  indicates the different derivative components (i.e. the derivative
                  with respect to the first parameter is [0,:,:], etc.)
        :rtype: ndarray
        """

        x1, n1, x2, n2, params, D = self._check_inputs(x1, x2, params)

        exp_theta = np.exp(-params[:(D - 1)])

        drdtheta = np.zeros((D - 1, n1, n2))

        r_matrix = self.calc_r(x1, x2, params)
        r_matrix[(r_matrix == 0.)] = 1.

        for d in range(D - 1):
            drdtheta[d] = (0.5 * np.exp(params[d]) / r_matrix *
                           cdist(np.reshape(x1[:,d], (n1, 1)),
                           np.reshape(x2[:,d], (n2, 1)), "sqeuclidean"))

        return drdtheta

    def calc_d2rdtheta2(self, x1, x2, params):
        r"""
        Calculate all second derivatives of the distance between all pairs of points with
        respect to the hyperparameters

        This method computes all second derivatives of the scaled Euclidean distance
        between all pairs of points in ``x1`` and ``x2`` with respect to the
        hyperparameters. The gradient is held in an array with shape ``(D, D, n1, n2)``,
        where ``D`` is the length of ``params``, ``n1`` is the length of the first axis
        of ``x1``, and ``n2`` is the length of the first axis of ``x2``. This is used in
        the computation of the gradient and Hessian of the kernel. The first two indices
        represents the different derivatives with respect to each hyperparameter.

        :param x1: First input array. Must be a 1-D or 2-D array, with the length of
                   the last dimension matching the last dimension of ``x2`` and
                   one less than the length of ``params``. ``x1`` may be 1-D if either
                   each point consists of a single parameter (and ``params`` has length
                   2) or the array only contains a single point (in which case, the array
                   will be reshaped to ``(1, D - 1)``).
        :type x1: array-like
        :param x2: Second input array. The same restrictions that apply to ``x1`` also
                   apply here.
        :type x2: array-like
        :param params: Hyperparameter array. Must be 1-D with length one greater than
                       the last dimension of ``x1`` and ``x2``.
        :type params: array-like
        :returns: Array holding the second derivatives of the pair-wise distances between
                  points in arrays ``x1`` and ``x2`` with respect to the hyperparameters.
                  Will be an array with shape ``(D, D, n1, n2)``, where ``D`` is the length
                  of ``params``, ``n1`` is the length of the first axis of ``x1`` and
                  ``n2`` is the length of the first axis of ``x2``. The first two axes
                  indicates the different derivative components (i.e. the second derivative
                  with respect to the first parameter is [0,0,:,:], the mixed partial with
                  respect to the first and second parameters is [0,1,:,:] or [1,0,:,:], etc.)
        :rtype: ndarray
        """

        x1, n1, x2, n2, params, D = self._check_inputs(x1, x2, params)

        exp_theta = np.exp(-params[:(D - 1)])

        d2rdtheta2 = np.zeros((D - 1, D - 1, n1, n2))

        r_matrix = self.calc_r(x1, x2, params)
        r_matrix[(r_matrix == 0.)] = 1.

        for d1 in range(D - 1):
            for d2 in range(D - 1):
                if d1 == d2:
                    d2rdtheta2[d1, d2] = (0.5*np.exp(params[d1]) / r_matrix *
                                          cdist(np.reshape(x1[:,d1], (n1, 1)),
                                                np.reshape(x2[:,d1], (n2, 1)), "sqeuclidean"))
                d2rdtheta2[d1, d2] -= (0.25 * np.exp(params[d1]) *
                                       np.exp(params[d2]) / r_matrix**3 *
                                       cdist(np.reshape(x1[:,d1], (n1, 1)),
                                             np.reshape(x2[:,d1], (n2, 1)), "sqeuclidean")*
                                       cdist(np.reshape(x1[:,d2], (n1, 1)),
                                             np.reshape(x2[:,d2], (n2, 1)), "sqeuclidean"))

        return d2rdtheta2

    def calc_drdx(self, x1, x2, params):
        r"""
        Calculate the first derivative of the distance between all pairs of points with
        respect to the first set of inputs

        This method computes the derivative of the scaled Euclidean distance between
        all pairs of points in ``x1`` and ``x2`` with respect to the first input ``x1``.
        The gradient is held in an array with shape ``(D - 1, n1, n2)``, where ``D`` is the
        length of ``params``, ``n1`` is the length of the first axis of
        ``x1``, and ``n2`` is the length of the first axis of ``x2``. This is used in the
        computation of the derivative of the kernel with respect to the inputs. The first
        index represents the different derivatives with respect to each input dimension.

        :param x1: First input array. Must be a 1-D or 2-D array, with the length of
                   the last dimension matching the last dimension of ``x2`` and
                   one less than the length of ``params``. ``x1`` may be 1-D if either
                   each point consists of a single parameter (and ``params`` has length
                   2) or the array only contains a single point (in which case, the array
                   will be reshaped to ``(1, D - 1)``).
        :type x1: array-like
        :param x2: Second input array. The same restrictions that apply to ``x1`` also
                   apply here.
        :type x2: array-like
        :param params: Hyperparameter array. Must be 1-D with length one greater than
                       the last dimension of ``x1`` and ``x2``.
        :type params: array-like
        :returns: Array holding the derivative of the pair-wise distances between
                  points in arrays ``x1`` and ``x2`` with respect to ``x1``.
                  Will be an array with shape ``(D, n1, n2)``, where ``D`` is the length
                  of ``params``, ``n1`` is the length of the first axis
                  of ``x1`` and ``n2`` is the length of the first axis of ``x2``. The first
                  axis indicates the different derivative components (i.e. the derivative
                  with respect to the first input parameter is [0,:,:], etc.)
        :rtype: ndarray
        """

        x1, n1, x2, n2, params, D = self._check_inputs(x1, x2, params)

        drdx = np.zeros((D - 1, n1, n2))

        exp_theta = np.exp(params[:(D - 1)])

        r_matrix = self.calc_r(x1, x2, params)
        r_matrix[(r_matrix == 0.)] = 1.

        for d in range(D - 1):
            drdx[d] = exp_theta[d]*(x1[:, d].flatten()[ :,    None ] -
                                    x2[:, d].flatten()[ None, :    ])/r_matrix

        return drdx

    def kernel_f(self, x1, x2, params):
        r"""
        Compute kernel values for a set of inputs

        Returns the value of the kernel for two sets of input points and a choice of
        hyperparameters. This function should not need to be modified for different choices
        of the kernel function or distance metric, as after checking the inputs it simply
        calls the routine to compute the distance metric and then evaluates the kernel function
        for those distances.

        :param x1: First input array. Must be a 1-D or 2-D array, with the length of
                   the last dimension matching the last dimension of ``x2`` and
                   one less than the length of ``params``. ``x1`` may be 1-D if either
                   each point consists of a single parameter (and ``params`` has length
                   2) or the array only contains a single point (in which case, the array
                   will be reshaped to ``(1, D - 1)``).
        :type x1: array-like
        :param x2: Second input array. The same restrictions that apply to ``x1`` also
                   apply here.
        :type x2: array-like
        :param params: Hyperparameter array. Must be 1-D with length one greater than
                       the last dimension of ``x1`` and ``x2``.
        :type params: array-like
        :returns: Array holding all kernel values between points in arrays ``x1``
                  and ``x2``. Will be an array with shape ``(n1, n2)``, where ``n1``
                  is the length of the first axis of ``x1`` and ``n2`` is the length
                  of the first axis of ``x2``.
        :rtype: ndarray
        """

        x1, n1, x2, n2, params, D = self._check_inputs(x1, x2, params)

        return np.exp(params[D - 1]) * self.calc_K(self.calc_r(x1, x2, params))

    def kernel_deriv(self, x1, x2, params):
        r"""
        Compute kernel gradient for a set of inputs

        Returns the value of the kernel gradient for two sets of input points and a choice of
        hyperparameters. This function should not need to be modified for different choices
        of the kernel function or distance metric, as after checking the inputs it simply
        calls the routine to compute the distance metric, kernel function, and the appropriate
        derivative functions of the distance and kernel functions.

        :param x1: First input array. Must be a 1-D or 2-D array, with the length of
                   the last dimension matching the last dimension of ``x2`` and
                   one less than the length of ``params``. ``x1`` may be 1-D if either
                   each point consists of a single parameter (and ``params`` has length
                   2) or the array only contains a single point (in which case, the array
                   will be reshaped to ``(1, D - 1)``).
        :type x1: array-like
        :param x2: Second input array. The same restrictions that apply to ``x1`` also
                   apply here.
        :type x2: array-like
        :param params: Hyperparameter array. Must be 1-D with length one greater than
                       the last dimension of ``x1`` and ``x2``.
        :type params: array-like
        :returns: Array holding the gradient of the kernel function between points in arrays
                  ``x1`` and ``x2`` with respect to the hyperparameters. Will be an array with
                  shape ``(D, n1, n2)``, where ``D`` is the length of ``params``, ``n1`` is the
                  length of the first axis of ``x1`` and ``n2`` is the length of the first axis
                  of ``x2``. The first axis indicates the different derivative components
                  (i.e. the derivative with respect to the first parameter is [0,:,:], etc.)
        :rtype: ndarray
        """

        x1, n1, x2, n2, params, D = self._check_inputs(x1, x2, params)

        dKdtheta = np.zeros((D, n1, n2))

        dKdtheta[-1] = self.kernel_f(x1, x2, params)

        dKdr = self.calc_dKdr(self.calc_r(x1, x2, params))

        drdtheta = self.calc_drdtheta(x1, x2, params)

        for d in range(D - 1):
            dKdtheta[d] = np.exp(params[-1]) * dKdr * drdtheta[d]

        return dKdtheta

    def kernel_hessian(self, x1, x2, params):
        r"""
        Calculate the Hessian of the kernel evaluated for all pairs of points with
        respect to the hyperparameters

        Returns the value of the kernel Hessian for two sets of input points and a choice of
        hyperparameters. This function should not need to be modified for different choices
        of the kernel function or distance metric, as after checking the inputs it simply
        calls the routine to compute the distance metric, kernel function, and the appropriate
        derivative functions of the distance and kernel functions.

        :param x1: First input array. Must be a 1-D or 2-D array, with the length of
                   the last dimension matching the last dimension of ``x2`` and
                   one less than the length of ``params``. ``x1`` may be 1-D if either
                   each point consists of a single parameter (and ``params`` has length
                   2) or the array only contains a single point (in which case, the array
                   will be reshaped to ``(1, D - 1)``).
        :type x1: array-like
        :param x2: Second input array. The same restrictions that apply to ``x1`` also
                   apply here.
        :type x2: array-like
        :param params: Hyperparameter array. Must be 1-D with length one greater than
                       the last dimension of ``x1`` and ``x2``.
        :type params: array-like
        :returns: Array holding the Hessian of the pair-wise distances between points in arrays
                  ``x1`` and ``x2`` with respect to the hyperparameters. Will be an array with
                  shape ``(D, D, n1, n2)``, where ``D`` is the length of ``params``, ``n1`` is
                  the length of the first axis of ``x1`` and ``n2`` is the length of the first
                  axis of ``x2``. The first two axes indicates the different derivative components
                  (i.e. the second derivative with respect to the first parameter is [0,0,:,:],
                  the mixed partial with respect to the first and second parameters is [0,1,:,:]
                  or [1,0,:,:], etc.)
        :rtype: ndarray
        """

        x1, n1, x2, n2, params, D = self._check_inputs(x1, x2, params)

        d2Kdtheta2 = np.zeros((D, D, n1, n2))

        d2Kdtheta2[-1, :] = self.kernel_deriv(x1, x2, params)
        d2Kdtheta2[:, -1] = d2Kdtheta2[-1, :]

        r_matrix = self.calc_r(x1, x2, params)
        dKdr = self.calc_dKdr(r_matrix)
        d2Kdr2 = self.calc_d2Kdr2(r_matrix)

        drdtheta = self.calc_drdtheta(x1, x2, params)
        d2rdtheta2 = self.calc_d2rdtheta2(x1, x2, params)

        for d1 in range(D - 1):
            for d2 in range(D - 1):
                d2Kdtheta2[d1, d2] = np.exp(params[-1]) * (d2Kdr2 *
                                                           drdtheta[d1] * drdtheta[d2] +
                                                           dKdr * d2rdtheta2[d1, d2])

        return d2Kdtheta2

    def kernel_inputderiv(self, x1, x2, params):
        r"""
        Compute derivative of Kernel with respect to inputs x1

        Returns the value of the kernel derivative with respect to the first set of input
        points given inputs and a choice of hyperparameters. This function should not need
        to be modified for different choices of the kernel function or distance metric, as
        after checking the inputs it simply calls the routine to compute the distance metric,
        kernel function, and the appropriate derivative functions of the distance and kernel
        functions.

        :param x1: First input array. Must be a 1-D or 2-D array, with the length of
                   the last dimension matching the last dimension of ``x2`` and
                   one less than the length of ``params``. ``x1`` may be 1-D if either
                   each point consists of a single parameter (and ``params`` has length
                   2) or the array only contains a single point (in which case, the array
                   will be reshaped to ``(1, D - 1)``).
        :type x1: array-like
        :param x2: Second input array. The same restrictions that apply to ``x1`` also
                   apply here.
        :type x2: array-like
        :param params: Hyperparameter array. Must be 1-D with length one greater than
                       the last dimension of ``x1`` and ``x2``.
        :type params: array-like
        :returns: Array holding the derivative of the kernel function between points in arrays
                  ``x1`` and ``x2`` with respect to the first inputs ``x1``. Will be an array with
                  shape ``(D, n1, n2)``, where ``D`` is the length of ``params``,
                  ``n1`` is the length of the first axis of ``x1`` and ``n2`` is the length of the
                  first axis of ``x2``. The first axis indicates the different derivative components
                  (i.e. the derivative with respect to the first input dimension is [0,:,:], etc.)
        :rtype: ndarray
        """

        x1, n1, x2, n2, params, D = self._check_inputs(x1, x2, params)

        dKdx = np.zeros((D - 1, n1, n2))

        r_matrix = self.calc_r(x1, x2, params)
        dKdr = self.calc_dKdr(r_matrix)

        drdx = self.calc_drdx(x1, x2, params)

        for d in range(D - 1):
            dKdx[d] = np.exp(params[-1]) * dKdr * drdx[d]

        return dKdx

    def calc_K(self, r):
        r"""
        Calculate kernel as a function of distance

        This method implements the kernel function as a function of distance. Given an array
        of distances, this function evaluates the kernel function of those values, returning
        an array of the same shape. Note that this is not implemented for the base class, as
        this must be defined for a specific kernel.

        :param r: Array holding distances between all points. All values in this array must be
                  non-negative.
        :type r: array-like
        :returns: Array holding kernel evaluations, with the same shape as the input ``r``
        :rtype: ndarray
        """

        raise NotImplementedError("base Kernel class does not implement a kernel function")

    def calc_dKdr(self, r):
        r"""
        Calculate first derivative of kernel as a function of distance

        This method implements the first derivative of the kernel function as a function of
        distance. Given an array of distances, this function evaluates the derivative
        function of those values, returning an array of the same shape. Note that this is
        not implemented for the base class, as this must be defined for a specific kernel.

        :param r: Array holding distances between all points. All values in this array must be
                  non-negative.
        :type r: array-like
        :returns: Array holding kernel derivatives, with the same shape as the input ``r``
        :rtype: ndarray
        """

        raise NotImplementedError("base Kernel class does not implement a kernel derivative function")

    def calc_d2Kdr2(self, r):
        r"""
        Calculate second derivative of kernel as a function of distance

        This method implements the second derivative of the kernel function as a function of
        distance. Given an array of distances, this function evaluates the second derivative
        function of those values, returning an array of the same shape. Note that this is
        not implemented for the base class, as this must be defined for a specific kernel.

        :param r: Array holding distances between all points. All values in this array must be
                  non-negative.
        :type r: array-like
        :returns: Array holding kernel second derivatives, with the same shape as the input ``r``
        :rtype: ndarray
        """

        raise NotImplementedError("base Kernel class does not implement kernel derivatives")

class SquaredExponential(Kernel):
    r"""
    Implementation of the squared exponential kernel

    Class representing a squared exponential kernel. It derives from the base class for a
    stationary kernel, using the scaled Euclidean distance metric. The subclass then just
    defines the kernel function and its derivatives.
    """

    def calc_K(self, r):
        r"""
        Compute K(r) for the squared exponential kernel

        This method implements the squared exponential kernel function as a function of distance.
        Given an array of distances, this function evaluates the kernel function of those values,
        returning an array of the same shape.

        :param r: Array holding distances between all points. All values in this array must be
                  non-negative.
        :type r: array-like
        :returns: Array holding kernel evaluations, with the same shape as the input ``r``
        :rtype: ndarray
        """

        assert np.all(r >= 0.), "kernel distances must be positive"

        r = np.array(r)

        return np.exp(-0.5*r**2)

    def calc_dKdr(self, r):
        r"""
        Calculate first derivative of the squared exponential kernel as a function of distance

        This method implements the first derivative of the squared exponential kernel function
        as a function of distance. Given an array of distances, this function evaluates the derivative
        function of those values, returning an array of the same shape.

        :param r: Array holding distances between all points. All values in this array must be
                  non-negative.
        :type r: array-like
        :returns: Array holding kernel derivatives, with the same shape as the input ``r``
        :rtype: ndarray
        """

        assert np.all(r >= 0.), "kernel distances must be positive"

        r = np.array(r)

        return -r*np.exp(-0.5*r**2)

    def calc_d2Kdr2(self, r):
        r"""
        Calculate second derivative of the squared exponential kernel as a function of distance

        This method implements the second derivative of the squared exponential kernel function
        as a function of distance. Given an array of distances, this function evaluates the
        second derivative function of those values, returning an array of the same shape.

        :param r: Array holding distances between all points. All values in this array must be
                  non-negative.
        :type r: array-like
        :returns: Array holding kernel second derivatives, with the same shape as the input ``r``
        :rtype: ndarray
        """

        assert np.all(r >= 0.), "kernel distances must be positive"

        r = np.array(r)

        return (r**2 - 1.)*np.exp(-0.5*r**2)

    def __str__(self):
        r"""
        Defines a string representation of the squared exponential kernel

        Returns a string representation of the squared exponential kernel. Note that since
        the kernel just collects methods for kernel evaluations together with no data, this
        is just a basic string that will not change for different instances of the class.

        :returns: String representation of the kernel
        :rtype: str
        """
        return "Squared Exponential Kernel"

class Matern52(Kernel):
    r"""
    Implementation of the Matern 5/2 kernel

    Class representing the Matern 5/2 kernel. It derives from the base class for a
    stationary kernel, using the scaled Euclidean distance metric. The subclass then just
    defines the kernel function and its derivatives.
    """
    def calc_K(self, r):
        r"""
        Compute K(r) for the Matern 5/2 kernel

        This method implements the Matern 5/2 kernel function as a function of distance.
        Given an array of distances, this function evaluates the kernel function of those values,
        returning an array of the same shape.

        :param r: Array holding distances between all points. All values in this array must be
                  non-negative.
        :type r: array-like
        :returns: Array holding kernel evaluations, with the same shape as the input ``r``
        :rtype: ndarray
        """

        assert np.all(r >= 0.), "kernel distances must be positive"

        r = np.array(r)

        return (1.+np.sqrt(5.)*r+5./3.*r**2)*np.exp(-np.sqrt(5.)*r)

    def calc_dKdr(self, r):
        r"""
        Calculate first derivative of the Matern 5/2 kernel as a function of distance

        This method implements the first derivative of the Matern 5/2 kernel function
        as a function of distance. Given an array of distances, this function evaluates the derivative
        function of those values, returning an array of the same shape.

        :param r: Array holding distances between all points. All values in this array must be
                  non-negative.
        :type r: array-like
        :returns: Array holding kernel derivatives, with the same shape as the input ``r``
        :rtype: ndarray
        """

        assert np.all(r >= 0.), "kernel distances must be positive"

        r = np.array(r)

        return -5./3.*r*(1.+np.sqrt(5.)*r)*np.exp(-np.sqrt(5.)*r)

    def calc_d2Kdr2(self, r):
        r"""
        Calculate second derivative of the squared exponential kernel as a function of distance

        This method implements the second derivative of the squared exponential kernel function
        as a function of distance. Given an array of distances, this function evaluates the
        second derivative function of those values, returning an array of the same shape.

        :param r: Array holding distances between all points. All values in this array must be
                  non-negative.
        :type r: array-like
        :returns: Array holding kernel second derivatives, with the same shape as the input ``r``
        :rtype: ndarray
        """

        assert np.all(r >= 0.), "kernel distances must be positive"

        r = np.array(r)

        return 5./3.*(5.*r**2-np.sqrt(5.)*r-1.)*np.exp(-np.sqrt(5.)*r)

    def __str__(self):
        r"""
        Defines a string representation of the Matern 5/2 kernel

        Returns a string representation of the Matern 5/2 kernel. Note that since
        the kernel just collects methods for kernel evaluations together with no data, this
        is just a basic string that will not change for different instances of the class.

        :returns: String representation of the kernel
        :rtype: str
        """
        return "Matern 5/2 Kernel"





"""
Simple demos and sanity checks for the HistoryMatching class.

Provided methods:
    get_y_simulated_1D:
        A toy model that acts as the simulator output for constructing GPEs for
        1D data.

    get_y_simulated_2D:
        A toy model that acts as the simulator output for constructing GPEs for
        2D data.

    demo_1D:
        Follows the example of
        http://www.int.washington.edu/talks/WorkShops/int_16_2a/People/Vernon_I/Vernon2.pdf
        in setting up a simple test model, constructing a gaussian process to
        emulate it, and ruling out expectations of the GPE based on an
        historical observation.

    demo_2D:
        As demo_1D, however the toy model is expanded to take two inputs rather
        then 1 extending it to a second dimension. Exists primarily to confirm
        that the HistoryMatching class is functional with higher-dimensional
        parameter spaces.

"""

from mogp_emulator.HistoryMatching import HistoryMatching
from mogp_emulator.GaussianProcess import GaussianProcess
from mogp_emulator.fitting import fit_GP_MAP
try:
    from matplotlib import pyplot as plt
    makeplots = True
except ImportError:
    makeplots = False
import numpy as np


def get_y_simulated_1D(x):
    n_points = len(x)
    f = np.zeros(n_points)
    for i in range(n_points):
        f[i] = np.sin(2.*np.pi*x[i] / 50.)
    return f

def demo_1D():
    # Create a gaussian process
    x_training = np.array([
        [0.],
        [10.],
        [20.],
        [30.],
        [43.],
        [50.]
    ])

    y_training = get_y_simulated_1D(x_training)

    gp = GaussianProcess(x_training, y_training)
    np.random.seed(47)
    gp = fit_GP_MAP(gp)

    # Define observation
    obs = [-0.8, 0.0004]

    # Coords to predict
    n_rand = 2000
    x_predict_min = -3
    x_predict_max = 53
    x_predict = np.random.rand(n_rand)
    x_predict = np.sort(x_predict, axis=0)
    x_predict *= (x_predict_max - x_predict_min)
    x_predict += x_predict_min
    x_predict = x_predict[:,None]

    coords = x_predict

    # Compute GPE expectations
    expectations = gp.predict(coords)

    # Compute Implausbility
    hm = HistoryMatching(obs=obs, expectations=expectations)
    I = hm.get_implausibility()
    NROY = hm.get_NROY()
    RO = hm.get_RO()

    print("Fraction of points ruled out {:6}".format(str(float(len(RO))/float(n_rand))))

    # Plotting

    if makeplots:
        fig, axs = plt.subplots(2, 1, sharex=True)
        fig.subplots_adjust(hspace=0)
        x_hist_plot = [min(x_predict)[0], max(x_predict)[0]]
        y_hist_plot = [obs[0], obs[0]]
        y_hist_err = 3*np.sqrt(obs[1])
        y_hist_up = [val + y_hist_err for val in y_hist_plot]
        y_hist_dn = [val - y_hist_err for val in y_hist_plot]

        axs[0].plot(                # Horizontal line at value of y_obs
            x_hist_plot,
            y_hist_plot,
            color = 'black',
            label = 'observation'
        )

        axs[0].fill_between(        # Error bounds on y_obs
            x_hist_plot,
            y_hist_dn,
            y_hist_up,
            color='black',
            alpha=0.25
        )

        axs[0].plot(                # Simulator output
            coords,
            get_y_simulated_1D(coords),
            color = 'black',
            label = 'simulator'
        )

        axs[0].scatter(             # Training Data
            x_training,
            y_training,
            marker = '.',
            color  = 'black',
            label  = 'Training Data',
            s      = 100
        )

        axs[0].plot(                # GPE expectation
            coords,
            expectations[0],
            color = 'red',
            label = 'GPE'
        )

        axs[0].fill_between(        # GPE uncertainty
            coords[:,0],
            expectations[0] - 3*np.sqrt(expectations[1]),
            expectations[0] + 3*np.sqrt(expectations[1]),
            color = 'red',
            alpha = 0.5
        )

        axs[1].scatter(             # Implausibility
            coords,
            I,
            marker='.',
            color='black'
        )

        axs[1].scatter(
            coords[NROY],
            I[NROY],
            marker='.',
            color='green'
        )

        axs[1].plot(                # implausibility Threshold
            x_hist_plot,
            [3,3],
            color = 'green',
            label = 'implausibility threshold'
        )

        axs[0].set(
            ylabel='Model Output f(x)'
        )
        axs[1].set(
            xlabel='Imput Parameter x',
            ylabel='Implausibility I(x)',
            ylim=(-1, 21)
        )


        plt.savefig('histmatch_1D.png', bbox_inches = 'tight')

def get_y_simulated_2D(x):
    n_points = len(x[:,0])
    f = np.zeros(n_points)
    for i in range(n_points):
        f[i] = np.sin(x[i,0] ** (2.0) + x[i,1] ** (2.0))
    return f

def demo_2D():
    # Create a Gaussian Process
    x_training = np.array([
        [0., 0.],
        [1.5, 1.5],
        [3., 3.],
        [0., 1.5],
        [1.5, 0.],
        [0., 3.],
        [3., 0.],
        [3., 1.5],
        [1.5, 3.]
    ])

    y_training = get_y_simulated_2D(x_training)

    gp = GaussianProcess(x_training, y_training)
    np.random.seed(47)
    gp = fit_GP_MAP(gp)

    # Define observation
    obs = [0.1, 0.0004]

    # Coords to predict
    n_rand = 2000
    a_predict_min = 0
    a_predict_max = np.pi
    b_predict_min = a_predict_min
    b_predict_max = a_predict_max
    a_predict = np.random.rand(n_rand) * (a_predict_max - a_predict_min) + a_predict_min
    b_predict = np.random.rand(n_rand) * (b_predict_max - b_predict_min) + b_predict_min
    x_predict = np.concatenate((a_predict[:,None], b_predict[:,None]), axis=1)
    x_predict = np.concatenate((x_predict, x_training), axis=0)

    coords = x_predict

    # Compute GPE expectations
    expectations = gp.predict(coords)

    # Compute Implausbility
    hm = HistoryMatching(obs=obs, expectations=expectations)
    I = hm.get_implausibility()
    NROY = hm.get_NROY()
    RO = hm.get_RO()

    print("Fraction of points ruled out {:6}".format(str(float(len(RO))/float(n_rand))))

    # Plotting
    if makeplots:
        from mpl_toolkits.mplot3d import Axes3D
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')

        Axes3D.scatter(     # Training Data
            ax,
            x_training[:,0],
            x_training[:,1],
            y_training,
            color='black',
            marker = '.',
            s = 100
        )

        #Axes3D.scatter(      # GPE prediction
        #    ax,
        #    coords[:,0],
        #    coords[:,1],
        #    expectations[0],
        #    color='red',
        #    marker='.',
        #    s=2
        #)

        Axes3D.scatter(     # GPE prediction uncertainty
            ax,
            coords[:,0][RO],
            coords[:,1][RO],
            expectations[0][RO] + 3*np.sqrt(expectations[1][RO]),
            color='red',
            marker='.',
            s=1
        )
        Axes3D.scatter(
            ax,
            coords[:,0][RO],
            coords[:,1][RO],
            expectations[0][RO] - 3*np.sqrt(expectations[1][RO]),
            color='red',
            marker='.',
            s=1
        )
        Axes3D.scatter(
            ax,
            coords[:,0][NROY],
            coords[:,1][NROY],
            expectations[0][NROY] + 3*np.sqrt(expectations[1][NROY]),
            color='green',
            marker='.',
            s=1
        )
        Axes3D.scatter(
            ax,
            coords[:,0][NROY],
            coords[:,1][NROY],
            expectations[0][NROY] - 3*np.sqrt(expectations[1][NROY]),
            color='green',
            marker='.',
            s=1
        )

        Axes3D.set(
            ax,
            xlabel = 'Input parameter a',
            ylabel = 'Input parameter b',
            zlabel = 'Model output f(a, b)'
        )



        plt.savefig('histmatch_2D.png', bbox_inches = "tight")


if __name__ == '__main__':
    demo_1D()
    demo_2D()
